/* -------------------------------------------- */
import { DarkStarsCombat } from "./dark-stars-combat.js";
import { DarkStarsCommands } from "./dark-stars-commands.js";

/* -------------------------------------------- */
const __locationNames = { head: "Head", chest: "Chest", abdomen: "Abdomen", leftarm: "Left Arm", rightarm: "Right Arm", leftleg: "Left Leg", rightleg: "Right Leg" }
/* -------------------------------------------- */
export class DarkStarsUtility {


  /* -------------------------------------------- */
  static async init() {
    Hooks.on('renderChatLog', (log, html, data) => DarkStarsUtility.chatListeners(html));
    /*Hooks.on("dropCanvasData", (canvas, data) => {
      DarkStarsUtility.dropItemOnToken(canvas, data)
    });*/

    DarkStarsCommands.init();

    Handlebars.registerHelper('count', function (list) {
      return list.length;
    })
    Handlebars.registerHelper('includes', function (array, val) {
      return array.includes(val);
    })
    Handlebars.registerHelper('upper', function (text) {
      return text.toUpperCase();
    })
    Handlebars.registerHelper('lower', function (text) {
      return text.toLowerCase()
    })
    Handlebars.registerHelper('upperFirst', function (text) {
      if (typeof text !== 'string') return text
      return text.charAt(0).toUpperCase() + text.slice(1)
    })
    Handlebars.registerHelper('notEmpty', function (list) {
      return list.length > 0;
    })
    Handlebars.registerHelper('mul', function (a, b) {
      return parseInt(a) * parseInt(b);
    })
    Handlebars.registerHelper('locationLabel', function (key) {
      return __locationNames[key]
    })
    

    this.gameSettings()

  }

  /*-------------------------------------------- */
  static gameSettings() {
    /*game.settings.register("fvtt-dark-stars", "dice-color-skill", {
      name: "Dice color for skills",
      hint: "Set the dice color for skills",
      scope: "world",
      config: true,
      requiresReload: true ,
      default: "#101010",
      type: String
    })

    Hooks.on('renderSettingsConfig', (event) => {
      const element = event.element[0].querySelector(`[name='fvtt-dark-stars.dice-color-skill']`)
      if (!element) return
      // Replace placeholder element
      console.log("Element Found !!!!")
    })    */
  }

  /*-------------------------------------------- */
  static upperFirst(text) {
    if (typeof text !== 'string') return text
    return text.charAt(0).toUpperCase() + text.slice(1)
  }

  /*-------------------------------------------- */
  static getSkills() {
    return duplicate(this.skills)
  }
  /*-------------------------------------------- */
  static getWeaponSkills() {
    return duplicate(this.weaponSkills)
  }
  /*-------------------------------------------- */
  static getShieldSkills() {
    return duplicate(this.shieldSkills)
  }

  /* -------------------------------------------- */
  static async ready() {
    const skills = await DarkStarsUtility.loadCompendium("fvtt-dark-stars.skills")
    this.skills = skills.map(i => i.toObject())
    this.weaponSkills = duplicate(this.skills.filter(item => item.system.isweaponskill))
    this.shieldSkills = duplicate(this.skills.filter(item => item.system.isshieldskill))

    const rollTables = await DarkStarsUtility.loadCompendium("fvtt-dark-stars.rolltables")
    this.rollTables = rollTables.map(i => i.toObject())

  }

  /* -------------------------------------------- */
  static async loadCompendiumData(compendium) {
    const pack = game.packs.get(compendium)
    return await pack?.getDocuments() ?? []
  }

  /* -------------------------------------------- */
  static async loadCompendium(compendium, filter = item => true) {
    let compendiumData = await DarkStarsUtility.loadCompendiumData(compendium)
    return compendiumData.filter(filter)
  }


  /* -------------------------------------------- */
  static async chatListeners(html) {

    html.on("click", '.view-item-from-chat', event => {
      game.system.crucible.creator.openItemView(event)
    })
    html.on("click", '.roll-defense-melee', event => {
      let rollId = $(event.currentTarget).data("roll-id")
      let rollData = DarkStarsUtility.getRollData(rollId)
      rollData.defenseWeaponId = $(event.currentTarget).data("defense-weapon-id")
      let actor = game.canvas.tokens.get(rollData.defenderTokenId).actor
      if (actor && (game.user.isGM || actor.isOwner)) {
        actor.rollDefenseMelee(rollData)
      }
    })
    html.on("click", '.roll-defense-ranged', event => {
      let rollId = $(event.currentTarget).data("roll-id")
      let rollData = DarkStarsUtility.getRollData(rollId)
      let defender = game.canvas.tokens.get(rollData.defenderTokenId).actor
      if (defender && (game.user.isGM || defender.isOwner)) {
        defender.rollDefenseRanged(rollData)
      }
    })

  }

  /* -------------------------------------------- */
  static async preloadHandlebarsTemplates() {

    const templatePaths = [
      'systems/fvtt-dark-stars/templates/partials/editor-notes-gm.hbs',
      'systems/fvtt-dark-stars/templates/partials/partial-roll-select.hbs',
      'systems/fvtt-dark-stars/templates/partials/partial-actor-ability-block.hbs',
      'systems/fvtt-dark-stars/templates/partials/partial-actor-status.hbs',
      'systems/fvtt-dark-stars/templates/partials/partial-options-abilities.hbs',
      'systems/fvtt-dark-stars/templates/partials/partial-item-nav.hbs',
      'systems/fvtt-dark-stars/templates/partials/partial-item-description.hbs',
      'systems/fvtt-dark-stars/templates/partials/partial-actor-equipment.hbs'
    ]
    return loadTemplates(templatePaths);
  }

  /* -------------------------------------------- */
  static removeChatMessageId(messageId) {
    if (messageId) {
      game.messages.get(messageId)?.delete();
    }
  }

  static findChatMessageId(current) {
    return DarkStarsUtility.getChatMessageId(DarkStarsUtility.findChatMessage(current));
  }

  static getChatMessageId(node) {
    return node?.attributes.getNamedItem('data-message-id')?.value;
  }

  static findChatMessage(current) {
    return DarkStarsUtility.findNodeMatching(current, it => it.classList.contains('chat-message') && it.attributes.getNamedItem('data-message-id'));
  }

  static findNodeMatching(current, predicate) {
    if (current) {
      if (predicate(current)) {
        return current;
      }
      return DarkStarsUtility.findNodeMatching(current.parentElement, predicate);
    }
    return undefined;
  }


  /* -------------------------------------------- */
  static createDirectOptionList(min, max) {
    let options = {};
    for (let i = min; i <= max; i++) {
      options[`${i}`] = `${i}`;
    }
    return options;
  }

  /* -------------------------------------------- */
  static buildListOptions(min, max) {
    let options = ""
    for (let i = min; i <= max; i++) {
      options += `<option value="${i}">${i}</option>`
    }
    return options;
  }

  /* -------------------------------------------- */
  static getTarget() {
    if (game.user.targets) {
      for (let target of game.user.targets) {
        return target
      }
    }
    return undefined
  }

  /* -------------------------------------------- */
  static async onSocketMesssage(msg) {
    console.log("SOCKET MESSAGE", msg.name)
    if (msg.name == "msg_update_roll") {
      this.updateRollData(msg.data)
    }
    if (msg.name == "msg_gm_process_attack_defense") {
      this.processSuccessResult(msg.data)
    }
    if (msg.name == "msg_gm_item_drop" && game.user.isGM) {
      let actor = game.actors.get(msg.data.actorId)
      let item
      if (msg.data.isPack) {
        item = await fromUuid("Compendium." + msg.data.isPack + "." + msg.data.itemId)
      } else {
        item = game.items.get(msg.data.itemId)
      }
      this.addItemDropToActor(actor, item)
    }
  }

  /* -------------------------------------------- */
  static chatDataSetup(content, modeOverride, isRoll = false, forceWhisper) {
    let chatData = {
      user: game.user.id,
      rollMode: modeOverride || game.settings.get("core", "rollMode"),
      content: content
    };

    if (["gmroll", "blindroll"].includes(chatData.rollMode)) chatData["whisper"] = ChatMessage.getWhisperRecipients("GM").map(u => u.id);
    if (chatData.rollMode === "blindroll") chatData["blind"] = true;
    else if (chatData.rollMode === "selfroll") chatData["whisper"] = [game.user];

    if (forceWhisper) { // Final force !
      chatData["speaker"] = ChatMessage.getSpeaker();
      chatData["whisper"] = ChatMessage.getWhisperRecipients(forceWhisper);
    }

    return chatData;
  }

  /* -------------------------------------------- */
  static async showDiceSoNice(roll, rollMode) {
    if (game.modules.get("dice-so-nice")?.active) {
      if (game.dice3d) {
        let whisper = null;
        let blind = false;
        rollMode = rollMode ?? game.settings.get("core", "rollMode");
        switch (rollMode) {
          case "blindroll": //GM only
            blind = true;
          case "gmroll": //GM + rolling player
            whisper = this.getUsers(user => user.isGM);
            break;
          case "roll": //everybody
            whisper = this.getUsers(user => user.active);
            break;
          case "selfroll":
            whisper = [game.user.id];
            break;
        }
        await game.dice3d.showForRoll(roll, game.user, true, whisper, blind);
      }
    }
  }

  /* -------------------------------------------- */
  static async rollDarkStars(rollData) {

    let actor = game.actors.get(rollData.actorId)

    // ability/save/size => 0
    let diceFormula
    let startFormula = "0d6cs>=5[blue]"
    if (rollData.ability) {
      startFormula = String(rollData.ability.value) + "d6cs>=5[blue]"
    }
    if (rollData.save) {
      startFormula = String(rollData.save.value) + "d6cs>=5[blue]"
    }
    if (rollData.sizeDice) {
      let nb = rollData.sizeDice.nb + rollData.distanceBonusDice + this.getDiceFromCover(rollData.hasCover) + this.getDiceFromSituational(rollData.situational)
      startFormula = String(nb) + String(rollData.sizeDice.dice) + "cs>=5[blue]"
    }
    diceFormula = startFormula

    // skill => 2
    // feat => 4
    // bonus => 6
    if (rollData.skill) {
      let level = rollData.skill.system.level
      if (rollData.skill.system.issl2) {
        rollData.hasSLBonus = true
        level += 2
        if (level > 7) { level = 7 }
      }
      rollData.skill.system.skilldice = __skillLevel2Dice[level]
      diceFormula += "+" + String(rollData.skill.system.skilldice) + "cs>=5[black]"

      if (rollData.skill.system.skilltype == "complex" && rollData.skill.system.level == 0) {
        rollData.complexSkillDisadvantage = true
        rollData.rollAdvantage = "roll-disadvantage"
      }

      if (rollData.skill.system.isfeatdie) {
        rollData.hasFeatDie = true
        diceFormula += "+ 1d10cs>=5[dark-stars-purple]"
      } else {
        diceFormula += `+ 0d10cs>=5[dark-stars-purple]`
      }
      if (rollData.skill.system.bonusdice != "none") {
        rollData.hasBonusDice = rollData.skill.system.bonusdice
        diceFormula += `+ ${rollData.hasBonusDice}cs>=5[black]`
      } else {
        diceFormula += `+ 0d6cs>=5[black]`
      }
    } else {
      diceFormula += `+ 0d8cs=>5 + 0d10cs>=5 + 0d6cs>=5`
    }

    // advantage => 8
    let advFormula = "+ 0d8cs>=5"
    if (rollData.advantage == "advantage1" || rollData.forceAdvantage) {
      advFormula = "+ 1d8cs>=5[dark-stars-darkgreen]"
    }
    if (rollData.advantage == "advantage2") {
      advFormula = "+ 2d8cs>=5[dark-stars-darkgreen]"
    }
    diceFormula += advFormula

    // disadvantage => 10
    let disFormula = "- 0d8cs>=5"
    if (rollData.disadvantage == "disadvantage1" || rollData.forceDisadvantage) {
      disFormula = "- 1d8cs>=5[red]"
    }
    if (rollData.disadvantage == "disadvantage2") {
      disFormula = "- 2d8cs>=5[red]"
    }
    diceFormula += disFormula

    // armor => 12
    let skillArmorPenalty = 0
    for (let armor of rollData.armors) {
      if (armor.system.equipped) {
        skillArmorPenalty += armor.system.skillpenalty
      }
    }
    if (rollData.skill && rollData.skill.system.armorpenalty && skillArmorPenalty > 0) {
      rollData.skillArmorPenalty = skillArmorPenalty
      diceFormula += `- ${skillArmorPenalty}d8cs>=5`
    } else {
      diceFormula += `- 0d8cs>=5`
    }

    // shield => 14
    if (rollData.useshield && rollData.shield) {
      diceFormula += "+ 1" + String(rollData.shield.system.shielddie) + "cs>=5[yellow]"
    } else {
      diceFormula += " + 0d6cs>=5"
    }

    // Performs roll
    console.log("Roll formula", diceFormula)
    let myRoll = rollData.roll
    if (!myRoll) { // New rolls only of no rerolls
      myRoll = new Roll(diceFormula).roll({ async: false })
      await this.showDiceSoNice(myRoll, game.settings.get("core", "rollMode"))
    }
    rollData.rollOrder = 0
    rollData.roll = myRoll
    rollData.nbSuccess = myRoll.total

    if (rollData.rollAdvantage == "none" && rollData.forceRollAdvantage) {
      rollData.rollAdvantage = "roll-advantage"
    }
    if (rollData.rollAdvantage == "none" && rollData.forceRollDisadvantage) {
      rollData.rollAdvantage = "roll-disadvantage"
    }
    if (rollData.rollAdvantage != "none") {

      rollData.rollOrder = 1
      rollData.rollType = (rollData.rollAdvantage == "roll-advantage") ? "Advantage" : "Disadvantage"
      this.createChatWithRollMode(rollData.alias, {
        content: await renderTemplate(`systems/fvtt-dark-stars/templates/chat-generic-result.hbs`, rollData)
      })

      rollData.rollOrder = 2
      let myRoll2 = new Roll(diceFormula).roll({ async: false })
      await this.showDiceSoNice(myRoll2, game.settings.get("core", "rollMode"))

      rollData.roll = myRoll2 // Tmp switch to display the proper results
      rollData.nbSuccess = myRoll2.total
      this.createChatWithRollMode(rollData.alias, {
        content: await renderTemplate(`systems/fvtt-dark-stars/templates/chat-generic-result.hbs`, rollData)
      })
      rollData.roll = myRoll // Revert the tmp switch
      rollData.nbSuccess = myRoll.total

      if (rollData.rollAdvantage == "roll-advantage") {
        if (myRoll2.total > rollData.nbSuccess) {
          hasChanged = true
          rollData.roll = myRoll2
          rollData.nbSuccess = myRoll2.total
        }
      } else {
        if (myRoll2.total < rollData.nbSuccess) {
          rollData.roll = myRoll2
          rollData.nbSuccess = myRoll2.total
        }
      }
      rollData.rollOrder = 3
    }
    rollData.nbSuccess = Math.max(0, rollData.nbSuccess)

    rollData.isFirstRollAdvantage = false
    // Manage exp
    if (rollData.skill && rollData.skill.system.level > 0) {
      let nbSkillSuccess = rollData.roll.terms[2].total
      if (nbSkillSuccess == 0 || nbSkillSuccess == rollData.skill.system.level) {
        actor.incrementSkillExp(rollData.skill.id, 1)
      }
    }

    this.saveRollData(rollData)
    actor.lastRoll = rollData

    this.createChatWithRollMode(rollData.alias, {
      content: await renderTemplate(`systems/fvtt-dark-stars/templates/chat-generic-result.hbs`, rollData)
    })
    console.log("Rolldata result", rollData)

    // Message response
    this.displayDefenseMessage(rollData)

    // Manage defense result
    this.processAttackDefense(rollData)
  }

  /* -------------------------------------------- */
  static sortArrayObjectsByName(myArray) {
    myArray.sort((a, b) => {
      let fa = a.name.toLowerCase();
      let fb = b.name.toLowerCase();
      if (fa < fb) {
        return -1;
      }
      if (fa > fb) {
        return 1;
      }
      return 0;
    })
  }

  /* -------------------------------------------- */
  static getUsers(filter) {
    return game.users.filter(filter).map(user => user.id);
  }
  /* -------------------------------------------- */
  static getWhisperRecipients(rollMode, name) {
    switch (rollMode) {
      case "blindroll": return this.getUsers(user => user.isGM);
      case "gmroll": return this.getWhisperRecipientsAndGMs(name);
      case "selfroll": return [game.user.id];
    }
    return undefined;
  }
  /* -------------------------------------------- */
  static getWhisperRecipientsAndGMs(name) {
    let recep1 = ChatMessage.getWhisperRecipients(name) || [];
    return recep1.concat(ChatMessage.getWhisperRecipients('GM'));
  }

  /* -------------------------------------------- */
  static blindMessageToGM(chatOptions) {
    let chatGM = duplicate(chatOptions);
    chatGM.whisper = this.getUsers(user => user.isGM);
    chatGM.content = "Blinde message of " + game.user.name + "<br>" + chatOptions.content;
    console.log("blindMessageToGM", chatGM);
    game.socket.emit("system.fvtt-dark-stars", { msg: "msg_gm_chat_message", data: chatGM });
  }


  /* -------------------------------------------- */
  static async searchItem(dataItem) {
    let item
    if (dataItem.pack) {
      item = await fromUuid("Compendium." + dataItem.pack + "." + dataItem.id)
    } else {
      item = game.items.get(dataItem.id)
    }
    return item
  }

  /* -------------------------------------------- */
  static split3Columns(data) {

    let array = [[], [], []];
    if (data == undefined) return array;

    let col = 0;
    for (let key in data) {
      let keyword = data[key];
      keyword.key = key; // Self-reference
      array[col].push(keyword);
      col++;
      if (col == 3) col = 0;
    }
    return array;
  }

  /* -------------------------------------------- */
  static createChatMessage(name, rollMode, chatOptions) {
    switch (rollMode) {
      case "blindroll": // GM only
        if (!game.user.isGM) {
          this.blindMessageToGM(chatOptions);

          chatOptions.whisper = [game.user.id];
          chatOptions.content = "Message only to the GM";
        }
        else {
          chatOptions.whisper = this.getUsers(user => user.isGM);
        }
        break;
      default:
        chatOptions.whisper = this.getWhisperRecipients(rollMode, name);
        break;
    }
    chatOptions.alias = chatOptions.alias || name;
    ChatMessage.create(chatOptions);
  }

  /* -------------------------------------------- */
  static getBasicRollData() {
    let rollData = {
      rollId: randomID(16),
      rollMode: game.settings.get("core", "rollMode"),
      advantage: "none"
    }
    DarkStarsUtility.updateWithTarget(rollData)
    return rollData
  }

  /* -------------------------------------------- */
  static updateWithTarget(rollData) {
    let target = DarkStarsUtility.getTarget()
    if (target) {
      rollData.defenderTokenId = target.id
    }
  }

  /* -------------------------------------------- */
  static createChatWithRollMode(name, chatOptions) {
    this.createChatMessage(name, game.settings.get("core", "rollMode"), chatOptions)
  }

  /* -------------------------------------------- */
  static async confirmDelete(actorSheet, li) {
    let itemId = li.data("item-id");
    let msgTxt = "<p>Are you sure to remove this Item ?";
    let buttons = {
      delete: {
        icon: '<i class="fas fa-check"></i>',
        label: "Yes, remove it",
        callback: () => {
          actorSheet.actor.deleteEmbeddedDocuments("Item", [itemId]);
          li.slideUp(200, () => actorSheet.render(false));
        }
      },
      cancel: {
        icon: '<i class="fas fa-times"></i>',
        label: "Cancel"
      }
    }
    msgTxt += "</p>";
    let d = new Dialog({
      title: "Confirm removal",
      content: msgTxt,
      buttons: buttons,
      default: "cancel"
    });
    d.render(true);
  }

}